/*****************************************************************************
    realpath.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2003 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <ctype.h>
#include <errno.h>

#include "params.h"
#include "globals.h"
#include "util.h"
#include "logfun.h"
#include "realpath.h"
    
/*
	check if <file> is in the directory tree below <base>
	this is done by resolving both <base> and <file> via realpath(2)
	and comparing the heads of the resulting paths
*/

int isInTree (char *base, char *file)
{
    char *rbase, *rfile;
    int rtc;

    if ((rbase = getPathBuf()) == NULL)
        return ENOMEM;

    if ((rfile = getPathBuf()) == NULL) {
        Free(rbase);
        return ENOMEM;
    }

    if ((realpath(base, rbase) != NULL) && (realpath(file, rfile) != NULL)) {
        if (!strncmp(rfile, rbase, strlen(rbase)))
            rtc = 0;
        else
            rtc = EACCES;
    }
    else
        rtc = errno; 

    Free(rbase);
    Free(rfile);
    return rtc;
}


int getPathMax(void)
{
#if (PATHMAX_MODE == BSD)
#define GetPathMax (MAXPATHLEN)
#elif (PATHMAX_MODE == SYS5)
#if defined PATH_MAX
#define GetPathMax (PATH_MAX)
#elif defined _POSIX_PATH_MAX
#define GetPathMax (_POSIX_PATH_MAX)
#endif
#endif
#ifndef GetPathMax
#error No maximum path length defined
#endif
    return (GetPathMax);
#undef GetPathMax
}


char *getPathBuf(void)
{
    return (char *)Calloc(1,getPathMax()+1);
}

/*
	add a path segment to a path buffer
	always prepends the segment with one slash
*/

 
static int addPBuf (char **act, char *str, int size)
{
    int l;

    if (size <= (l = strlen(str)+1))	/* we need strlen()+2 free space */	
        return ENAMETOOLONG;

    *(*act) = '/';			/* apeend one slash */
    strcpy (((*act)+1), str);		/* plus the new path segment */

    *act  += l;				/* adjust the pointer */

    return 0;
}


static int softLink(char *buf, int flags)
{
    struct stat sb;

    if (flags & RP_KEEPLINKS)
        return 0;

    if (lstat(buf, &sb))
        return 0;

    return S_ISLNK(sb.st_mode);
}

/*
	take a root path and a partial (relative) path,
        concatenate them and remove self-relative constructs ( . and ..)

	path:	input path string
	buf:	output buffer (resolved path)
	act:	pointer into <buf> := active write position
	size:	remaining size in <buf>
	flags:	RP_KEEPDOTS	don't resolve . and ..
		RP_KEEPLINKS	don't resolve soft links
		RP_TILDE	resolve ~
		RP_NODOLLAR	consider $ a syntax error
		RP_CTS		reslove context sensitive links (HP Tru64 only)
*/
static int realPath2(char *path, char *buf, char **act, int size, int flags)
{
    struct stat sb;
    char *oldact, *ctx=NULL, *tok, *temp, *sep;
    int rtc, bad=0, len;

    if (path == NULL)
        return 0;
/*
	split the given path into chunks	
*/
    for (tok=xstrtok(path,"/",&ctx); tok!=NULL; tok=xstrtok(NULL,"/",&ctx)) {
        if (bad)
            bad = ENOTDIR;

        if (!strcmp(tok, "."))		/* self relative to current dir	*/
            continue;			/* simply ignore		*/

	if (!(flags & RP_KEEPDOTS) &&
            !strcmp(tok, "..")) {	/* self relative to parent dir	*/
            if ((sep = strrchr(buf, '/')) != NULL) /* got parent dir ? 	*/
                *act = sep;		/* yes, remove it !		*/
            *sep = '\0';		/* terminate buffer		*/
            if ((bad = stat(buf, &sb)))	/* valid partial path ?		*/
                bad = errno;		
            continue;
        }

	if ((flags & RP_TILDE) &&
            !strcmp(tok, "~")) {	/* home-dir expansion		*/
            if ((temp = getenv("HOME")) != NULL) {
                if ((temp = Strdup(temp)) == NULL)
                    return errno;
                if (*temp == '/')		/* absolute path ?	*/
                    *act = buf;			/* yes, install new path*/

                rtc = realPath2(temp, buf, act, size-(int)(*act-buf), flags); 
                Free(temp);			/* discard HOME		*/

                if (rtc) {			/* link invalid !	*/
                    if ((rtc == ENOENT) || (rtc == ENOTDIR))
                        bad = rtc;		/* continue with error	*/
                    else
                        return rtc;		/* severe, quit		*/
                }
            }
            continue;
        }
/*
	here on regular chunks
*/
        if ((flags & RP_NODOLLAR) &&
            (strchr(tok, '$') != NULL))
            return EINVAL;

        oldact  = *act;			/* save active position		*/
					/* and store this chunk		*/

        if ((rtc = addPBuf(act, tok, size - (int)(*act - buf))))
            return rtc;

/*
	if no error occured yet, let's test the so far path construct
*/
        if (!bad) {
            if (softLink(buf, flags)) {			/* soft link	*/
                if ((temp = getPathBuf()) == NULL)	/* get a buffer	*/
                    return ENOMEM;
						/* resolve the link 	*/

                if ((len = readlink(buf, temp, getPathMax())) < 0) {
                    Free(temp);				/* failed	*/
                    return errno;
                }
                *(temp+len) = '\0';		/* terminate 		*/
                if (*temp == '/')		/* absolute link ?	*/
                    *act = buf;			/* yes, install new path*/
                else {				/* no			*/
                    *oldact = '\0';		/* get rid of link	*/
                    *act  = oldact;		/* and restore old path	*/
                }
/*	have ourselves check the chunks appended from the soft link	*/

                rtc = realPath2(temp, buf, act, size-(int)(*act-buf), flags); 
                Free(temp);			/* discard the link	*/

                if (rtc) {			/* link invalid !	*/
                    if ((rtc == ENOENT) || (rtc == ENOTDIR))
                        bad = rtc;		/* continue with error	*/
                    else
                        return rtc;		/* severe, quit		*/
                }
            }
            else {			/* regular chunk (no soft link)	*/
                if (stat(buf, &sb))
                    bad = errno;
            }
        }
    }
    return bad;
}

/*
	jacket routine for realPath2, which does the real work

	concatenates <root> and <relpath> and evaluates the resulting path:
	self-relative constructs and soft links are resolved by default.

	flags:	RP_KEEPDOTS	don't resolve . and ..
		RP_KEEPLINKS	don't resolve soft links
		RP_TILDE	resolve ~
		RP_NODOLLAR	consider $ a syntax error
		RP_CTS		reslove context sensitive links (HP Tru64 only)

	other than realpath(2) this one doesn't stop resolving if one of
	the errors ENOENT or ENODIR occures, but returns a complete path
	anyway. in this way, path constructs leading to not yet created
	files may be checked and resolved.
*/
int realPath (char *root, char *relpath, char **resolvedpath, int flags) 
{
    int  rtc = 0, size;
    char *cwd, *buf, *act, *temp;

    if ((root == NULL) && (relpath == NULL))	/* need at least one 	*/
        return EINVAL;				/* partial path		*/

    *resolvedpath = NULL;			/* in spite of errors	*/
    size = getPathMax();			/* size of result-buffer*/

    if ((buf = act = getPathBuf()) == NULL) {	/* get the buffer itself*/
        errNoDynmem;
        return ENOMEM;
    }

    if (root    == NULL) root    = "";
    if (relpath == NULL) relpath = "";

/*
	prepend the path specifications with the current working dir
	if needed to assure an absolute path
*/
    if (((*root != '\0') && (*root != '/')) ||
        ((*root == '\0') && (*relpath != '/'))) {
        if ((temp = getPathBuf()) == NULL) {
            Free(buf);
            errNoDynmem;
            return ENOMEM;
        }
        if ((cwd = getcwd(temp, getPathMax())) == NULL) {
            Free(buf);
            Free(temp);
            return errno;
        }
        rtc = realPath2(cwd, buf, &act, size - (int)(act-buf), flags);
        Free(temp);
    }

    if (!rtc || (rtc == ENOENT) || (rtc == ENOTDIR)) {
        if (*root != '\0') {
            if ((temp = Strdup(root)) == NULL) {
                Free(buf);
                errNoDynmem;
                return ENOMEM;
            }
            rtc = realPath2(temp, buf, &act, size - (int)(act-buf), flags);
            Free(temp);
        }
    }

    if (!rtc || (rtc == ENOENT) || (rtc == ENOTDIR)) {
        if (*relpath != '\0') {
            if ((temp = Strdup(relpath)) == NULL) {
                Free(buf);
                errNoDynmem;
                return ENOMEM;
            }
            rtc = realPath2(temp, buf, &act, size - (int)(act-buf), flags);
            Free(temp);
        }
    }

    *resolvedpath = buf;

    return rtc;
}

#ifdef TEST
int main (int argc, char **argv)
{
    char *root=NULL, *relpath=NULL, *result=NULL;
    int ios, flags=0;

    if (argc < 3) {
        fprintf (stderr, "Usage is: root relpath [flags]\n");
        return 1;
    }

    if (argc > 1) root=argv[1];
    if (argc > 2) relpath=argv[2];
    if (argc > 3) flags=atoi(argv[3]);

    ios = realPath(root, relpath, &result, flags);

    printf ("%s: ios: %s\n", __FUNCTION__, strerror(ios));
    if (result != NULL)
        puts (result);

    return ios;
}
#endif /* TEST */   
